@@ -19,29 +19,36 @@ use Net::WebSocket::Frame::close ();
 use Net::WebSocket::Frame::ping ();
 use Net::WebSocket::Frame::pong ();
 use Net::WebSocket::Message ();
+use Net::WebSocket::PingStore ();
 use Net::WebSocket::X ();
 
-use constant DEFAULT_MAX_PINGS => 2;
+use constant DEFAULT_MAX_PINGS => 3;
 
 sub new {
     my ($class, %opts) = @_;
 
-    my @missing = grep { !length $opts{$_} } qw( parser out );
-    #die "Missing: [@missing]" if @missing;
+    my @missing = grep { !length $opts{$_} } qw( out parser );
+
+    die "Missing: [@missing]" if @missing;
+
+    if ( !(ref $opts{'out'})->can('write') ) {
+        die "“out” ($opts{'out'}) needs a write() method!";
+    }
 
     my $self = {
         _fragments => [],
 
-        _ping_counter => 0,
         _max_pings => $class->DEFAULT_MAX_PINGS(),
 
+        _ping_store => Net::WebSocket::PingStore->new(),
+
         (map { defined($opts{$_}) ? ( "_$_" => $opts{$_} ) : () } qw(
             parser
-            out
             max_pings
 
             on_data_frame
-            before_send_control_frame
+
+            out
         )),
     };
 
@@ -51,45 +58,151 @@ sub new {
 sub get_next_message {
     my ($self) = @_;
 
-    die "Already closed!" if $self->{'_closed'};
+    $self->_verify_not_closed();
 
-    if ( my $frame = $self->{'_parser'}->get_next_frame() ) {
-        if ($frame->is_control_frame()) {
-            $self->_handle_control_frame($frame);
+    my $_msg_frame;
+
+    if ( $_msg_frame = $self->{'_parser'}->get_next_frame() ) {
+        if ($_msg_frame->is_control_frame()) {
+            $self->_handle_control_frame($_msg_frame);
         }
         else {
             if ($self->{'_on_data_frame'}) {
-                $self->{'_on_data_frame'}->($frame);
+                $self->{'_on_data_frame'}->($_msg_frame);
             }
 
             #Failure cases:
             #   - continuation without prior fragment
             #   - non-continuation within fragment
 
-            if ( $frame->get_type() eq 'continuation' ) {
+            if ( $_msg_frame->get_type() eq 'continuation' ) {
                 if ( !@{ $self->{'_fragments'} } ) {
-                    $self->_got_continuation_during_non_fragment($frame);
+                    $self->_got_continuation_during_non_fragment($_msg_frame);
                 }
             }
             elsif ( @{ $self->{'_fragments'} } ) {
-                $self->_got_non_continuation_during_fragment($frame);
+                $self->_got_non_continuation_during_fragment($_msg_frame);
             }
 
-            if ($frame->get_fin()) {
+            if ($_msg_frame->get_fin()) {
                 return Net::WebSocket::Message::create_from_frames(
                     splice( @{ $self->{'_fragments'} } ),
-                    $frame,
+                    $_msg_frame,
                 );
             }
             else {
-                push @{ $self->{'_fragments'} }, $frame;
+                push @{ $self->{'_fragments'} }, $_msg_frame;
             }
         }
     }
 
-    return undef;
+    return defined($_msg_frame) ? q<> : undef;
 }
 
+sub check_heartbeat {
+    my ($self) = @_;
+
+    my $ping_counter = $self->{'_ping_store'}->get_count();
+
+    if ($ping_counter == $self->{'_max_pings'}) {
+        $self->close(
+            code => 'POLICY_VIOLATION',
+            reason => "Unanswered ping(s): $ping_counter",
+        );
+    }
+
+    my $ping_message = $self->{'_ping_store'}->add();
+
+    my $ping = Net::WebSocket::Frame::ping->new(
+        payload_sr => \$ping_message,
+        $self->FRAME_MASK_ARGS(),
+    );
+
+    $self->_write_frame($ping);
+
+    return;
+}
+
+sub close {
+    my ($self, %opts) = @_;
+
+    my $close = Net::WebSocket::Frame::close->new(
+        $self->FRAME_MASK_ARGS(),
+        code => $opts{'code'} || 'ENDPOINT_UNAVAILABLE',
+        reason => $opts{'reason'},
+    );
+
+    return $self->_close_with_frame($close);
+}
+
+sub _close_with_frame {
+    my ($self, $close_frame) = @_;
+
+    $self->_write_frame($close_frame);
+
+    $self->{'_sent_close_frame'} = $close_frame;
+
+    return $self;
+}
+
+*shutdown = *close;
+
+sub is_closed {
+    my ($self) = @_;
+    return $self->{'_sent_close_frame'} ? 1 : 0;
+}
+
+sub received_close_frame {
+    my ($self) = @_;
+    return $self->{'_received_close_frame'};
+}
+
+sub sent_close_frame {
+    my ($self) = @_;
+    return $self->{'_sent_close_frame'};
+}
+
+sub die_on_close {
+    my ($self) = @_;
+
+    $self->{'_no_die_on_close'} = 0;
+
+    return;
+}
+
+sub do_not_die_on_close {
+    my ($self) = @_;
+
+    $self->{'_no_die_on_close'} = 1;
+
+    return;
+}
+
+#----------------------------------------------------------------------
+
+sub on_ping {
+    my ($self, $frame) = @_;
+
+    $self->_write_frame(
+        Net::WebSocket::Frame::pong->new(
+            payload_sr => \$frame->get_payload(),
+            $self->FRAME_MASK_ARGS(),
+        ),
+    );
+
+    return;
+}
+
+sub on_pong {
+    my ($self, $frame) = @_;
+
+    $self->{'_ping_store'}->remove( $frame->get_payload() );
+
+    return;
+}
+
+#----------------------------------------------------------------------
+
 sub _got_continuation_during_non_fragment {
     my ($self, $frame) = @_;
 
@@ -104,7 +217,7 @@ sub _got_continuation_during_non_fragment {
         $self->FRAME_MASK_ARGS(),
     );
 
-    $self->_send_frame($err_frame);
+    $self->_write_frame($err_frame);
 
     die Net::WebSocket::X->create( 'ReceivedBadControlFrame', $msg );
 }
@@ -123,97 +236,39 @@ sub _got_non_continuation_during_fragment {
         $self->FRAME_MASK_ARGS(),
     );
 
-    $self->_send_frame($err_frame);
+    $self->_write_frame($err_frame);
 
     die Net::WebSocket::X->create( 'ReceivedBadControlFrame', $msg );
 }
 
-sub check_heartbeat {
+sub _verify_not_closed {
     my ($self) = @_;
 
-    if ($self->{'_ping_counter'} == $self->{'_max_pings'}) {
-        my $close = Net::WebSocket::Frame::close->new(
-            code => 'POLICY_VIOLATION',
-            $self->FRAME_MASK_ARGS(),
-        );
-        print { $self->{'_out'} } $close->to_bytes();
-        $self->{'_closed'} = 1;
-    }
-
-    my $ping_message = sprintf("%s UTC: $self->{'_ping_counter'} (%s)", scalar(gmtime), rand);
-
-    $self->{'_ping_texts'}{$ping_message} = 1;
-
-    my $ping = Net::WebSocket::Frame::ping->new(
-        payload_sr => \$ping_message,
-        $self->FRAME_MASK_ARGS(),
-    );
-    print { $self->{'_out'} } $ping->to_bytes();
-
-    $self->{'_ping_counter'}++;
+    die Net::WebSocket::X->create('EndpointAlreadyClosed') if $self->{'_closed'};
 
     return;
 }
 
-sub is_closed {
-    my ($self) = @_;
-    return $self->{'_closed'} ? 1 : 0;
-}
-
-#----------------------------------------------------------------------
-
-sub on_ping {
-    my ($self, $frame) = @_;
-
-    $self->_send_frame(
-        Net::WebSocket::Frame::pong->new(
-            payload_sr => \$frame->get_payload(),
-            $self->FRAME_MASK_ARGS(),
-        ),
-    );
-
-    return;
-}
-
-sub on_pong {
-    my ($self, $frame) = @_;
-
-    #NB: We expect a response to any ping that we’ve sent; any pong
-    #we receive that doesn’t actually correlate to a ping we’ve sent
-    #is ignored—i.e., it doesn’t reset the ping counter. This means that
-    #we could still timeout even if we’re receiving pongs.
-    if (delete $self->{'_ping_texts'}{$frame->get_payload()}) {
-        $self->{'_ping_counter'} = 0;
-    }
-
-    return;
-}
-
-#----------------------------------------------------------------------
-
 sub _handle_control_frame {
     my ($self, $frame) = @_;
 
-    my ($resp_frame, $error, $ignore_sigpipe);
-
     my $type = $frame->get_type();
 
     if ($type eq 'close') {
-        $self->{'_closed'} = 1;
+        if (!$self->{'_sent_close_frame'}) {
+            $self->_close_with_frame($frame);
+        }
 
-        my ($code, $reason) = $frame->get_code_and_reason();
+        if ($self->{'_received_close_frame'}) {
+            warn sprintf('Extra close frame received! (%v.02x)', $frame->to_bytes());
+        }
+        else {
+            $self->{'_received_close_frame'} = $frame;
+        }
 
-        $resp_frame = Net::WebSocket::Frame::close->new(
-            code => $code,
-            reason => $reason,
-            $self->FRAME_MASK_ARGS(),
-        );
-
-        local $SIG{'PIPE'} = 'IGNORE';
-
-        $self->_send_frame($resp_frame);
-
-        die Net::WebSocket::X->create('ReceivedClose', $frame);
+        if (!$self->{'_no_die_on_close'}) {
+            die Net::WebSocket::X->create('ReceivedClose', $frame);
+        }
     }
     elsif ( my $handler_cr = $self->can("on_$type") ) {
         $handler_cr->( $self, $frame );
@@ -229,16 +284,10 @@ sub _handle_control_frame {
     return;
 }
 
-sub _send_frame {
+sub _write_frame {
     my ($self, $frame) = @_;
 
-    if ($self->can('_before_send_frame')) {
-        $self->_before_send_frame($frame);
-    }
-
-    print { $self->{'_out'} } $frame->to_bytes();
-
-    return;
+    return $self->{'_out'}->write($frame->to_bytes());
 }
 
 1;
